@@ -1,6 +1,6 @@
 /*
  * cups - Python bindings for CUPS
- * Copyright (C) 2002, 2005, 2006, 2007  Tim Waugh <twaugh@redhat.com>
+ * Copyright (C) 2002, 2005, 2006, 2007, 2008, 2009  Tim Waugh <twaugh@redhat.com>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -25,53 +25,58 @@
 #include "cupsmodule.h"
 
 #include <locale.h>
+#include <wchar.h>
+#include <wctype.h>
 
 #include "cupsconnection.h"
 #include "cupsppd.h"
 
 static PyObject *cups_password_callback = NULL;
+void *g_current_connection = NULL;
 
 //////////////////////
 // Worker functions //
 //////////////////////
 
 static int
-do_model_compare (const char *a, const char *b)
+do_model_compare (const wchar_t *a, const wchar_t *b)
 {
-  const char *digits = "0123456789";
-  char quick_a, quick_b;
-  while ((quick_a = *a) != '\0' && (quick_b = *b) != '\0') {
-    int end_a = strspn (a, digits);
-    int end_b = strspn (b, digits);
+  const wchar_t *digits = L"0123456789";
+  wchar_t quick_a, quick_b;
+  while ((quick_a = *a) != L'\0' && (quick_b = *b) != L'\0') {
+    int end_a = wcsspn (a, digits);
+    int end_b = wcsspn (b, digits);
     int min;
     int a_is_digit = 1;
     int cmp;
 
-    if (quick_a != quick_b && !isdigit (quick_a) && !isdigit (quick_b)) {
+    if (quick_a != quick_b && !iswdigit (quick_a) && !iswdigit (quick_b)) {
       if (quick_a < quick_b) return -1;
       else return 1;
     }
 
     if (!end_a) {
-      end_a = strcspn (a, digits);
+      end_a = wcscspn (a, digits);
       a_is_digit = 0;
     }
 
     if (!end_b) {
       if (a_is_digit)
 	return -1;
-      end_b = strcspn (b, digits);
+      end_b = wcscspn (b, digits);
     } else if (!a_is_digit)
       return 1;
 
     if (a_is_digit) {
-      int n_a = atoi (a), n_b = atoi (b);
+      unsigned long n_a, n_b;
+      n_a = wcstoul (a, NULL, 10);
+      n_b = wcstoul (b, NULL, 10);
       if (n_a < n_b) cmp = -1;
       else if (n_a == n_b) cmp = 0;
       else cmp = 1;
     } else {
       min = end_a < end_b ? end_a : end_b;
-      cmp = strncmp (a, b, min);
+      cmp = wcsncmp (a, b, min);
     }
 
     if (!cmp) {
@@ -86,8 +91,8 @@ do_model_compare (const char *a, const char *b)
     return cmp;
   }
 
-  if (quick_a == '\0') {
-    if (*b == '\0')
+  if (quick_a == L'\0') {
+    if (*b == L'\0')
       return 0;
 
     return -1;
@@ -105,11 +110,17 @@ do_password_callback (const char *prompt)
   PyObject *result;
   const char *pwval;
 
+  debugprintf ("-> do_password_callback\n");
+  Connection_end_allow_threads (g_current_connection);
   args = Py_BuildValue ("(s)", prompt);
   result = PyEval_CallObject (cups_password_callback, args);
   Py_DECREF (args);
   if (result == NULL)
+  {
+    debugprintf ("<- do_password_callback (empty string)\n");
+    Connection_begin_allow_threads (g_current_connection);
     return "";
+  }
 
   if (password) {
     free (password);
@@ -120,8 +131,14 @@ do_password_callback (const char *prompt)
   password = strdup (pwval);
   Py_DECREF (result);
   if (!password)
+  {
+    debugprintf ("<- do_password_callback (empty string)\n");
+    Connection_begin_allow_threads (g_current_connection);
     return "";
-  
+  }
+
+  Connection_begin_allow_threads (g_current_connection);
+  debugprintf ("<- do_password_callback\n");
   return password;
 }
 
@@ -132,12 +149,64 @@ do_password_callback (const char *prompt)
 static PyObject *
 cups_modelSort (PyObject *self, PyObject *args)
 {
-  char *a, *b;
+  PyObject *Oa, *Ob, *a, *b;
+  int len_a, len_b;
+  size_t size_a, size_b;
+  wchar_t *wca, *wcb;
 
-  if (!PyArg_ParseTuple (args, "ss", &a, &b))
+  if (!PyArg_ParseTuple (args, "OO", &Oa, &Ob))
     return NULL;
 
-  return Py_BuildValue ("i", do_model_compare (a, b));
+  a = PyUnicode_FromObject (Oa);
+  b = PyUnicode_FromObject (Ob);
+  if (a == NULL || b == NULL || !PyUnicode_Check (a) || !PyUnicode_Check (b)) {
+    if (a) {
+      Py_DECREF (a);
+    }
+    if (b) {
+      Py_DECREF (b);
+    }
+
+    PyErr_SetString (PyExc_TypeError, "Unable to convert to Unicode");
+    return NULL;
+  }
+
+  len_a = 1 + PyUnicode_GetSize (a);
+  size_a = len_a * sizeof (wchar_t);
+  if ((size_a / sizeof (wchar_t)) != len_a) {
+    Py_DECREF (a);
+    Py_DECREF (b);
+    PyErr_SetString (PyExc_RuntimeError, "String too long");
+    return NULL;
+  }
+
+  wca = malloc (size_a);
+
+  len_b = 1 + PyUnicode_GetSize (b);
+  size_b = len_b * sizeof (wchar_t);
+  if ((size_b / sizeof (wchar_t)) != len_b) {
+    Py_DECREF (a);
+    Py_DECREF (b);
+    PyErr_SetString (PyExc_RuntimeError, "String too long");
+    return NULL;
+  }
+
+  wcb = malloc (size_b);
+  
+  if (wca == NULL || wcb == NULL) {
+    Py_DECREF (a);
+    Py_DECREF (b);
+    free (wca);
+    free (wcb);
+    PyErr_SetString (PyExc_RuntimeError, "Insufficient memory");
+    return NULL;
+  }
+
+  PyUnicode_AsWideChar ((PyUnicodeObject *) a, wca, size_a);
+  PyUnicode_AsWideChar ((PyUnicodeObject *) b, wcb, size_b);
+  Py_DECREF (a);
+  Py_DECREF (b);
+  return Py_BuildValue ("i", do_model_compare (wca, wcb));
 }
 
 static PyObject *
@@ -236,6 +305,18 @@ cups_setPasswordCB (PyObject *self, PyObject *args)
   return Py_None;
 }
 
+static PyObject *
+cups_ppdSetConformance (PyObject *self, PyObject *args)
+{
+  int level;
+  if (!PyArg_ParseTuple (args, "i", &level))
+    return NULL;
+
+  ppdSetConformance (level);
+  Py_INCREF (Py_None);
+  return Py_None;
+}
+
 static PyObject *
 cups_require (PyObject *self, PyObject *args)
 {
@@ -253,13 +334,18 @@ cups_require (PyObject *self, PyObject *args)
   nreq = strtoul (preq, &end, 0);
   while (preq != end)
   {
-    preq = end + 1;
+    preq = end;
+    if (*preq == '.')
+      preq++;
 
     nver = strtoul (pver, &end, 0);
     if (pver == end)
       goto fail;
-    else
-      pver = end + 1;
+    else {
+      pver = end;
+      if (*pver == '.')
+	pver++;
+    }
 
     if (nver < nreq)
       goto fail;
@@ -334,6 +420,12 @@ static PyMethodDef CupsMethods[] = {
     "@type fn: callable object\n"
     "@param fn: callback function" },
 
+  { "ppdSetConformance", cups_ppdSetConformance, METH_VARARGS,
+    "ppdSetConformance(level) -> None\n\n"
+    "Set PPD conformance level.\n\n"
+    "@type level: integer\n"
+    "@param level: PPD_CONFORM_RELAXED or PPD_CONFORM_STRICT" },
+
   { "require", cups_require, METH_VARARGS,
     "require(version) -> None\n\n"
     "Require pycups version.\n\n"
@@ -441,6 +533,9 @@ initcups (void)
   INT_CONSTANT (CUPS_PRINTER_AUTHENTICATED);
   INT_CONSTANT (CUPS_PRINTER_COMMANDS);
   INT_CONSTANT (CUPS_PRINTER_OPTIONS);
+#if CUPS_VERSION_MAJOR > 1 || (CUPS_VERSION_MAJOR == 1 && CUPS_VERSION_MINOR >= 3)
+  INT_CONSTANT (CUPS_PRINTER_DISCOVERED);
+#endif /* CUPS 1.3 */
 
   // HTTP encryption
   INT_CONSTANT (HTTP_ENCRYPT_IF_REQUESTED);
@@ -557,13 +652,17 @@ initcups (void)
   // Limits
   INT_CONSTANT (IPP_MAX_NAME);
 
+  // PPD conformance levels
+  INT_CONSTANT (PPD_CONFORM_RELAXED);
+  INT_CONSTANT (PPD_CONFORM_STRICT);
+
   // Admin Util constants
   STR_CONSTANT (CUPS_SERVER_DEBUG_LOGGING);
   STR_CONSTANT (CUPS_SERVER_REMOTE_ADMIN);
   STR_CONSTANT (CUPS_SERVER_REMOTE_PRINTERS);
   STR_CONSTANT (CUPS_SERVER_SHARE_PRINTERS);
   STR_CONSTANT (CUPS_SERVER_USER_CANCEL_ANY);
-#ifdef HAVE_CUPS_1_3
+#if CUPS_VERSION_MAJOR > 1 || (CUPS_VERSION_MAJOR == 1 && CUPS_VERSION_MINOR >= 3)
   STR_CONSTANT (CUPS_SERVER_REMOTE_ANY);
 #endif /* CUPS 1.3 */
 
